/*
 * Copyright 2012 The Apache Software Foundation.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.odftoolkit.odfdom.component;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.nio.charset.Charset;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.json.JSONArray;
import org.json.JSONObject;
import org.json.JSONTokener;
import org.odftoolkit.odfdom.doc.OdfDocument;
import org.odftoolkit.odfdom.doc.OdfSpreadsheetDocument;
import org.odftoolkit.odfdom.doc.OdfTextDocument;
import org.odftoolkit.odfdom.dom.OdfSchemaDocument.OdfXMLFile;
import org.odftoolkit.odfdom.pkg.OdfPackage;

/**
 * This is a prototype of a new high level view text document. Methods might be
 * better bIt will be added
 *
 * @author svante.schubertATgmail.com
 */
public class OdfOperationDocument {

	private static final Logger LOG = Logger.getLogger(OdfOperationDocument.class.getName());
	// not private as used as well by tests
	static final String OPERATION_REVISON_FILE = "debug/revision.txt";
	static final String OPERATION_TEXT_FILE_PREFIX = "debug/operationUpdates_";
	static final String ORIGNAL_ODT_FILE = "debug/original.odt";
	/**
	 * Only being used when a certain test (my workbench test, which is not part
	 * of the regression testing) is being used
	 */
	private static final String OPERATION_DEBUG_OUTPUT_FILE = "../../../testy/odf-operations.txt";
	private OdfDocument mDocument;
	private OdfPackage mPackage;
	private Map<Long, byte[]> mResourceMap;
	// The language of the document
	private String mUserLanguage;
	private boolean mSaveDebugOperations;
	private int mMaxTableColumnCount;
	private int mMaxTableRowCount;
	private int mMaxTableCellCount;
	private boolean isMetadataUpdated = false;

	/**
	 * Creates an empty ODF document.
	 */
	private OdfOperationDocument() {
	}

	/**
	 * Creates a new ODT document from the default template
	 */
	public static OdfOperationDocument newTextOperationDocument() throws Exception {
		OdfOperationDocument odt = new OdfOperationDocument();
		odt.mDocument = OdfTextDocument.newTextDocument();
		return odt;
	}

	/**
	 * Creates a new ODS document from the default template
	 */
	public static OdfOperationDocument newSpreadsheetOperationDocument() throws Exception {
		OdfOperationDocument ods = new OdfOperationDocument();
		ods.mDocument = OdfSpreadsheetDocument.newSpreadsheetDocument();
		return ods;
	}

	/**
	 * Creates an OdfOperationDocument from the OpenDocument provided by a
	 * resource Stream.
	 *
	 * <p>Since an InputStream does not provide the arbitrary (non sequential)
	 * read access needed by OdfOperationDocument, the InputStream is cached.
	 * This usually takes more time compared to the other createInternalDocument
	 * methods. An advantage of caching is that there are no problems
	 * overwriting an input file.</p>
	 *
	 * @param inputStream - the InputStream of the ODF text document.
	 * @return the text document created from the given InputStream
	 */
	public OdfOperationDocument(InputStream aInputDocumentStm) throws Exception {
		mDocument = OdfDocument.loadDocument(aInputDocumentStm);
	}

	/**
	 * Creates an OdfOperationDocument from the OpenDocument provided by a
	 * resource Stream.
	 *
	 * <p>Since an InputStream does not provide the arbitrary (non sequential)
	 * read access needed by OdfOperationDocument, the InputStream is cached.
	 * This usually takes more time compared to the other createInternalDocument
	 * methods. An advantage of caching is that there are no problems
	 * overwriting an input file.</p>
	 *
	 * @param inputStream - the InputStream of the ODF text document.
	 * @param saveDebugOperations - the operations and original file will be
	 * stored in the output ODF document
	 * @return the text document created from the given InputStream
	 */
	public OdfOperationDocument(InputStream aInputDocumentStm, Map<String, Object> configuration) throws Exception {
		this(aInputDocumentStm, null, configuration);
	}

	/**
	 * Creates an OdfOperationDocument from the OpenDocument provided by a
	 * resource Stream.
	 *
	 * <p>Since an InputStream does not provide the arbitrary (non sequential)
	 * read access needed by OdfOperationDocument, the InputStream is cached.
	 * This usually takes more time compared to the other createInternalDocument
	 * methods. An advantage of caching is that there are no problems
	 * overwriting an input file.</p>
	 *
	 * @param inputStream - the InputStream of the ODF text document.
	 * @param resourceMap - the bytes of new resources can be accessed by an ID.
	 * @param configuration - key/value pairs of user given run-time settings
	 * (configuration)
	 * @return the text document created from the given InputStream
	 */
	public OdfOperationDocument(InputStream inputDocumentStm, Map<Long, byte[]> resourceMap, Map<String, Object> configuration) throws Exception {
		mDocument = OdfDocument.loadDocument(inputDocumentStm, configuration);
		mPackage = getDocument().getPackage();
		mResourceMap = resourceMap;
		if (configuration != null) {
			if (configuration.containsKey("debugoperations")) {
				mSaveDebugOperations = (Boolean) configuration.get("debugoperations");
			}
			if (configuration.containsKey("maxTableColumns")) {
				mMaxTableColumnCount = (Integer) configuration.get("maxTableColumns");
			}
			if (configuration.containsKey("maxTableRows")) {
				mMaxTableRowCount = (Integer) configuration.get("maxTableRows");
			}
			if (configuration.containsKey("maxTableCells")) {
				mMaxTableCellCount = (Integer) configuration.get("maxTableCells");
			}
		}

	}

	/**
	 * Creates an OdfOperationDocument from the OpenDocument provided by a
	 * resource Stream.
	 *
	 * <p>Since an InputStream does not provide the arbitrary (non sequential)
	 * read access needed by OdfOperationDocument, the InputStream is cached.
	 * This usually takes more time compared to the other createInternalDocument
	 * methods. An advantage of caching is that there are no problems
	 * overwriting an input file.</p>
	 *
	 * @param inputStream - the InputStream of the ODF text document.
	 * @param resourceMap - the bytes of new resources can be accessed by an ID.
	 * @param userLanguage - the language of the document
	 * @param configuration - key/value pairs of user given run-time settings
	 * (configuration)
	 * @return the text document created from the given InputStream
	 */
	public OdfOperationDocument(InputStream inputDocumentStm, Map<Long, byte[]> resourceMap, String userLanguage, Map<String, Object> configuration) throws Exception {
		this(inputDocumentStm, resourceMap, configuration);
		mUserLanguage = userLanguage;
	}

	public int getMaxTableColumnsCount() {
		return mMaxTableColumnCount;
	}

	public int getMaxTableRowsCount() {
		return mMaxTableRowCount;
	}

	public int getMaxTableCellCount() {
		return mMaxTableCellCount;
	}

	/**
	 * Receives the a map with new resources for the Document
	 *
	 * @return the operations as JSON
	 */
	public Map<Long, byte[]> getResourceMap() {
		return mResourceMap;
	}

	/**
	 * Receives the (known) operations of the ODF text document
	 *
	 * @return the operations as JSON
	 */
	// ToDo OX - JSONObject is to be considered..
	public JSONObject getOperations() {
		JSONObject ops = mDocument.getOperations();
		if (ops != null) {
			LOG.log(Level.FINER, "\n\n*** ALL OPERATIONS:\n{0}", mDocument.getOperations().toString());
		} else {
			LOG.finer("\n\n*** ALL OPERATIONS:\nNo Operation could be extracted!");
		}
		return ops;
	}

	/**
	 * Applies the (known) operations to upon the latest state of the ODF text
	 * document
	 *
	 * @param operationString ODF operations as String
	 * @return the number of operations being accepted
	 */
	public int applyOperations(String operationString) {
		JSONObject operations = new JSONObject();
		try {
			operations.put("operations", new JSONArray(new JSONTokener(operationString)));
		} catch (Exception ex) {
			LOG.log(Level.SEVERE, null, ex);
		}
		int operationsCount = this.applyOperations(operations);
		// if the document was altered
		if (operationsCount > 0) {
			// remove the cached view
			removeCachedView();
		}
		return operationsCount;
	}

	/**
	 * Applies the (known) operations to upon the latest state of the ODF text
	 * document
	 *
	 * @param operations ODF operations as JSONArray within an JSONObject with
	 * "operations" key.
	 * @return the number of operations being accepted
	 */
	public int applyOperations(JSONObject operations) {
		LOG.log(Level.FINER, "\n*** EDIT OPERATIONS:\n{0}", operations.toString());
//		System.err.println("\n*** EDIT OPERATIONS:\n{0}" + operations.toString());
		int operationsCount = 0;
		try {
			final JSONArray ops = operations.getJSONArray("operations");
			if (mSaveDebugOperations) {
				addOriginalOdfAsDebug();
				addOperationFileAsDebug(ops);
			}
			saveLocalDebug(ops);
			operationsCount = JsonOperationConsumer.applyOperations(this, ops);
		} catch (Exception ex) {
			LOG.log(Level.SEVERE, null, ex);
		}
		// if the document was altered
		if (operationsCount > 0) {
			// remove the cached view
			removeCachedView();
			if (!isMetadataUpdated) {
				mDocument.updateMetaData();
				isMetadataUpdated = true;
			}
		}
		return operationsCount;
	}

	private static void saveLocalDebug(JSONArray ops) {
		// only meant for local testing
		String unitTest = System.getProperty("test");
		if (unitTest != null && unitTest.equals("org.odftoolkit.odfdom.component.LatestOngoingTest")) {
			saveOperationAsDebugFile(ops, OPERATION_DEBUG_OUTPUT_FILE);
		}
	}

	private void removeCachedView() {
		if (mPackage == null) {
			mPackage = getDocument().getPackage();
		}
		// removes the LO/AO view caching
		mPackage.remove("Thumbnails/thumbnail.png");
	}

	private void addOriginalOdfAsDebug() {
		OdfPackage pkg = mDocument.getPackage();
		// if there is not already an orignal file being stored
		if (!pkg.contains(ORIGNAL_ODT_FILE)) {
			LOG.finest("Adding original ODT document as debug within the zip at " + ORIGNAL_ODT_FILE);
			try {
				// ..from the ODF ZIP 
				pkg.insert(pkg.getInputStream(), ORIGNAL_ODT_FILE, "application/vnd.oasis.opendocument.text");
			} catch (IOException ex) {
				Logger.getLogger(OdfOperationDocument.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
	}

	/*
	 * @param file the file to be saved, when creating a test file, you might use <code>newTestOutputFile(String relativeFilePath)</code>.
	 * @param inputData the data to be written into the file
	 */
	private static void saveStringToFile(File file, String data) {
		saveStringToFile(file, Charset.forName("UTF-8"), data);
	}

	/**
	 * @param file the file to be saved, when creating a test file, you might
	 * use <code>newTestOutputFile(String relativeFilePath)</code>.
	 * @param charset the character encoding
	 * @param inputData the data to be written into the file
	 */
	private static void saveStringToFile(File file, Charset charset, String inputData) {
		BufferedWriter out = null;
		try {
			out = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), charset));
			//out = new BufferedWriter(new FileWriter(file));
			out.write(inputData);
			out.close();
		} catch (IOException ex) {
			Logger.getLogger(OdfOperationDocument.class.getName()).log(Level.SEVERE, null, ex);
		} finally {
			try {
				out.close();
			} catch (IOException ex) {
				Logger.getLogger(OdfOperationDocument.class.getName()).log(Level.SEVERE, null, ex);
			}
		}
	}

	private static void saveOperationAsDebugFile(JSONArray operations, String debugFilePath) {
		// serialize the operations as String (using ascii characters only) and indent a line for every new operations (heuristic: every array item will be split into new line)
		String indentOps = JsonOperationNormalizer.asString(operations).replace("},{", "},\n{");
		if (debugFilePath != null && !debugFilePath.isEmpty()) {
			saveStringToFile(new File(debugFilePath), indentOps);
		}
	}

	private void addOperationFileAsDebug(JSONArray operations) {
		// serialize the operations as String (using ascii characters only) and indent a line for every new operations (heuristic: every array item will be split into new line)
		String indentOps = JsonOperationNormalizer.asString(operations).replace("},{", "},\n{");
		try {
			OdfPackage pkg = mDocument.getPackage();
			// start with zero to always increment (either read a default by file or new)
			int revisionNo = 0;
			// if there was already a revision, get it..
			if (pkg.contains(OPERATION_REVISON_FILE)) {
				// ..from the ODF ZIP 
				byte[] revisionByteArray = pkg.getBytes(OPERATION_REVISON_FILE);
				if (revisionByteArray != null && revisionByteArray.length != 0) {
					BufferedReader reader = new BufferedReader(new InputStreamReader(new ByteArrayInputStream(revisionByteArray)));
					// read the first line of the file, only containing one number
					String firstLine = reader.readLine();
					// map it to a number
					revisionNo = Integer.parseInt(firstLine);
					LOG.log(Level.FINE, "Found an existing revision number:{0}", revisionNo);
				}
			} else {
				LOG.fine("Created a new revision number: 1");
			}
			// always increment, so even a new file starts with the revision number "1"
			revisionNo++;
			pkg.insert(indentOps.getBytes(), OPERATION_TEXT_FILE_PREFIX + revisionNo + ".txt", "text/plain");
			pkg.insert(Integer.toString(revisionNo).getBytes(), OPERATION_REVISON_FILE, "text/plain");
		} catch (Exception ex) {
			Logger.getLogger(OdfOperationDocument.class.getName()).log(Level.SEVERE, null, ex);
		}
	}

	public long getContentSize() {
		if (mPackage == null) {
			if (mDocument == null) {
			}
			mPackage = mDocument.getPackage();
		}
		if (mPackage != null) {
			return mPackage.getSize(OdfXMLFile.CONTENT.getFileName());
		} else {
			return 0;
		}
	}

	/**
	 * Returns the OdfTextDocument encapsulating the DOM view
	 *
	 * @return ODF text document
	 */
	public OdfDocument getDocument() {
		return mDocument;
	}

	/**
	 * Close the OdfPackage and release all temporary created data. After
	 * execution of this method, this class is no longer usable. Do this as the
	 * last action to free resources. Closing an already closed document has no
	 * effect.
	 */
	public void close() {
		mDocument.close();
		mResourceMap = null;
	}
}
